using System;
using System.Collections.Generic;
using System.IO;
using RJH.CommandLineHelper;

namespace SqlBi.DtsToSsisPrepare {
    public class PackageMigration {
        public class MigrationParameters {
            private bool _consolidateUdl = false;
            private bool _optimizeConnections = false;
            private bool _removeAdvancedTransformations = false;
            private bool _resetMaxErrorCount = false;
            private string _dtsSourcePath;
            private string _dtsDestinationPath;

            [CommandLineSwitch( "ConsolidateUdl", "Consolidate UDL connections into regular connections")]
            public bool ConsolidateUdl {
                get { return this._consolidateUdl; }
                set { this._consolidateUdl = value; }
            }

            [CommandLineSwitch( "OptimizeConnections", "Optimize connections recycling the same connection for several tasks" )]
            public bool OptimizeConnections {
                get { return this._optimizeConnections; }
                set { this._optimizeConnections = value; }
            }

            [CommandLineSwitch( "RemoveAdvancedTransformations", "Remove advanced transformations phases from Transform Data Task" )]
            public bool RemoveAdvancedTransformations {
                get { return this._removeAdvancedTransformations; }
                set { this._removeAdvancedTransformations = value; }
            }

            [CommandLineSwitch( "ResetMaxErrorCount", "Reset Maximum Error Count property from Transform Data Task" )]
            public bool ResetMaxErrorCount {
                get { return this._resetMaxErrorCount; }
                set { this._resetMaxErrorCount = value; }
            }

            [CommandLineSwitch( "DtsSource", "Source DTS file name" )]
            public string DtsSourcePath {
                get { return this._dtsSourcePath; }
                set { this._dtsSourcePath = value; }
            }

            [CommandLineSwitch( "DtsDestination", "Destination DTS file name" )]
            public string DtsDestinationPath {
                get { return this._dtsDestinationPath; }
                set { this._dtsDestinationPath = value; }
            }

            
            public void ResetAll() {
                ConsolidateUdl = false;
                OptimizeConnections = false;
                RemoveAdvancedTransformations = false;
                ResetMaxErrorCount = false;
            }
            
            public void SetAll() {
                ConsolidateUdl = true;
                OptimizeConnections = true;
                RemoveAdvancedTransformations = true;
                ResetMaxErrorCount = true;
            }
            public bool IsOneActive() {
                return ConsolidateUdl || OptimizeConnections || RemoveAdvancedTransformations || ResetMaxErrorCount;
            }
        }

        public MigrationParameters Parameters = new MigrationParameters();
        
        private object host = null;
        private DTS.Package2 package;

        public void Migrate( string dtsInputPath, string dtsOutputPath ) {
            OpenFile( dtsInputPath );
            MigrateInternal();
            SaveFile( dtsOutputPath );
        }

        public bool ArgumentCheck() {
            if (!File.Exists( Parameters.DtsSourcePath )) {
                Log( MigrationMessage.Error, String.Format( "Input file {0} not found", Parameters.DtsSourcePath ) );
                return false;
            }

            if (!Parameters.IsOneActive()) {
                Log( MigrationMessage.Error, String.Format( "You have to set at least one conversion switch (suggested to enable all)" ) );
                return false;
            }
            return true;
        }
        public void Migrate() {
            if (!ArgumentCheck()) {
                return;
            }
            
            Migrate( Parameters.DtsSourcePath, Parameters.DtsDestinationPath );
        }
        
        public PackageMigration() {
            package = (DTS.Package2) new DTS.Package();
        }
        public PackageMigration( string filename ) : this() {
            OpenFile( filename );
        }

        public void OpenFile( string path ) {
            this.package = (DTS.Package2) new DTS.Package();
            this.package.LoadFromStorageFile( path, null, null, null, null, ref this.host );
        }

        public void SaveFile( string path ) {
            if (File.Exists( path )) {
                File.Delete( path );
            }
            this.package.SaveToStorageFileAs( package.Name, path, null, null, ref this.host, false );
        }
        
        protected void MigrateInternal() {
            if (Parameters.OptimizeConnections) {
                // Optimize connections and consolidate UDLs
                OptimizeConnections();
                PurgeUnreferencedConnections();
            }
            else if (Parameters.ConsolidateUdl) {
                // Consolidate UDLs only if no optimize connections required: inside OptimizeConnections
                // the UDLs consolidation is automatically done
                ConsolidateUdlConnections();
            }
            CleanTransformationTask();
        }

        /// <summary>
        /// Clean Data Transformation Task changing properties incompatibles with SSIS migration
        /// </summary>
        public void CleanTransformationTask() {
            foreach (DTS.Task task in package.Tasks) {
                if (task.CustomTask is DTS.DataPumpTask2) {
                    DTS.DataPumpTask2 dataTask = (DTS.DataPumpTask2) task.CustomTask;
                    // It is not required to reset exception file name - it will be ignored anyway - dataTask.ExceptionFileName = "";
                    if (Parameters.ResetMaxErrorCount) {
                        dataTask.MaximumErrorCount = 0;
                    }
                    if (Parameters.RemoveAdvancedTransformations) {
                        RemoveTransformationAdvancedPhases( dataTask );    
                    }
                }
            }
        }

        /// <summary>
        /// Remove transformation of advanced phases - the presence of an advanced phase can't be migrated to SSIS
        /// </summary>
        /// <param name="dataTask">Data Task on which operate</param>
        public void RemoveTransformationAdvancedPhases( DTS.DataPumpTask2 dataTask ) {
            for (int i = dataTask.Transformations.Count; i > 0; i-- ) {
                DTS.Transformation2 transformation = (DTS.Transformation2) dataTask.Transformations.Item( i );
                if (transformation.TransformPhases != 4) {
                    Log( MigrationMessage.Warning,
                         String.Format( "task {0} - removed advanced transformation phase", dataTask.Name ) );
                    dataTask.Transformations.Remove( i );
                }
            }
        }

        private Dictionary<string, DTS.Connection2> _optimizedConnections = new Dictionary<string, DTS.Connection2>();
        private Dictionary<int, DTS.Connection2> _mapConnections = new Dictionary<int, DTS.Connection2>();
        
        /// <summary>
        /// Create a dictionary of optimized connections, consolidating UDL connections
        /// </summary>
        /// <param name="existingConnection"></param>
        /// <returns></returns>
        public DTS.Connection2 GetOptimizedConnection( DTS.Connection2 existingConnection ) {
            DTS.Connection2 optimizedConnection;
            if ((existingConnection.UDLPath != null) && (existingConnection.UDLPath.Length > 0)) {
                // Read UDL coniguration
                IniFile iniFile = new IniFile( existingConnection.UDLPath );
                string connectionString = iniFile.ReadValue( "oledb", "Provider" );

                // Parse connection string
                ConnectionInfo connectionInfo = new ConnectionInfo();
                connectionInfo.Parse( connectionString );

                string key = String.Format( "{0}-{1}", connectionInfo.DataSource, connectionInfo.InitialCatalog );
                if (!_optimizedConnections.TryGetValue( key, out optimizedConnection )) {
                    optimizedConnection = CreateOptimizedConnection( connectionInfo, existingConnection, key );
                }
            }
            else {
                string key = String.Format( "{0}-{1}-{2}", existingConnection.DataSource, existingConnection.Catalog, existingConnection.ProviderID );
                if (!_optimizedConnections.TryGetValue( key, out optimizedConnection )) {
                    optimizedConnection = existingConnection;
                    _optimizedConnections.Add( key, optimizedConnection );
                }
            }
            return optimizedConnection;
        }

        // TODO: Initialize with max id found in existing connections
        private int _connectionIdCounter = 1000;
        
        private DTS.Connection2 CreateOptimizedConnection( ConnectionInfo connectionInfo, DTS.Connection2 existingConnection, string key ) {
            DTS.Connection2 optimizedConnection;
            optimizedConnection = (DTS.Connection2) this.package.Connections.New( "SQLOLEDB" );
            ConsolidateUdlConnection( connectionInfo, existingConnection, optimizedConnection, _connectionIdCounter++, key );
            this._optimizedConnections.Add( key, optimizedConnection );

            package.Connections.Add( (DTS.Connection) optimizedConnection );
            
            return optimizedConnection;
        }

        /// <summary>
        /// Optimize Connections
        /// </summary>
        public void OptimizeConnections() {
            foreach (DTS.Connection2 existingConnection in package.Connections) {
                if ((existingConnection.UDLPath != null) && (existingConnection.UDLPath.Length > 0)) {
                    DTS.Connection2 optimizedConnection = GetOptimizedConnection( existingConnection );
                    _mapConnections.Add( existingConnection.ID, optimizedConnection );
                }
            }
            
            // Loop into tasks to optimize
            foreach (DTS.Task task in package.Tasks) {
                if (task.CustomTask is DTS.ExecuteSQLTask2) {
                    DTS.ExecuteSQLTask2 sqlTask = (DTS.ExecuteSQLTask2) task.CustomTask;
                    DTS.Connection2 optimizedConnection;
                    if (_mapConnections.TryGetValue( sqlTask.ConnectionID, out optimizedConnection )) {
                        // Change connection ID
                        sqlTask.ConnectionID = optimizedConnection.ID;
                    }
                    else {
                        Log( MigrationMessage.Warning, 
                             String.Format( "{0} has connection ID={1} that can't be optimized", task.Name, sqlTask.ConnectionID ));
                    }
                }
                else if (task.CustomTask is DTS.DataPumpTask2) {
                    DTS.DataPumpTask2 dataTask = (DTS.DataPumpTask2) task.CustomTask;
                    DTS.Connection2 optimizedConnection;
                    if (_mapConnections.TryGetValue( dataTask.SourceConnectionID, out optimizedConnection )) {
                        // Change connection ID
                        dataTask.SourceConnectionID = optimizedConnection.ID;
                    }
                    else {
                        Log( MigrationMessage.Warning,
                             String.Format( "{0} has connection ID={1} that can't be optimized", task.Name, dataTask.SourceConnectionID ) );
                    }
                    if (_mapConnections.TryGetValue( dataTask.DestinationConnectionID, out optimizedConnection )) {
                        // Change connection ID
                        dataTask.DestinationConnectionID = optimizedConnection.ID;
                    }
                    else {
                        Log( MigrationMessage.Warning,
                             String.Format( "{0} has connection ID={1} that can't be optimized", task.Name, dataTask.DestinationConnectionID ) );
                    }
                }
            }

        }
        
        /// <summary>
        /// Transform UDL connections in regular connections without a global optimization
        /// </summary>
        public void ConsolidateUdlConnections() {
            for (int i = package.Connections.Count; i > 0; i--) {
                DTS.Connection2 oldConnection = (DTS.Connection2) package.Connections.Item( i );
                if ((oldConnection.UDLPath != null) && (oldConnection.UDLPath.Length > 0)) {
                    DTS.Connection2 newConnection = (DTS.Connection2) package.Connections.New( "SQLOLEDB" );
                    newConnection.ID = oldConnection.ID;
                    ConsolidateUdlConnection( oldConnection, newConnection );
                    package.Connections.Remove( i );
                    package.Connections.Add( (DTS.Connection) newConnection );
                }
            }
        }

        /// <summary>
        /// Purge unused connections
        /// </summary>
        public void PurgeUnreferencedConnections() {
            for (int indexConnection = package.Connections.Count; indexConnection > 0; indexConnection--) {
                DTS.Connection2 connection = (DTS.Connection2) package.Connections.Item( indexConnection );
                int connectionID = connection.ID;
                bool used = false;
                foreach (DTS.Task task in package.Tasks) {
                    if (task.CustomTask is DTS.DataPumpTask) {
                        DTS.DataPumpTask dataTask = (DTS.DataPumpTask) task.CustomTask;
                        if ((dataTask.DestinationConnectionID == connectionID) || (dataTask.SourceConnectionID == connectionID)) {
                            used = true;
                            break;
                        }
                    }
                    else if (task.CustomTask is DTS.ExecuteSQLTask) {
                        DTS.ExecuteSQLTask sqlTask = (DTS.ExecuteSQLTask) task.CustomTask;
                        if (sqlTask.ConnectionID == connectionID) {
                            used = true;
                            break;
                        }
                    }
                }
                if (!used) {
                    package.Connections.Remove( indexConnection );
                }
            }
        }
        
        /// <summary>
        /// Transform a UDL connection in a regular connection
        /// </summary>
        /// <param name="oldConnection">Old connection to consolidate</param>
        /// <param name="newConnection">New connection (copy from old and consolidate)</param>
        protected void ConsolidateUdlConnection( DTS.Connection2 oldConnection, DTS.Connection2 newConnection ) {
            // Read UDL coniguration
            IniFile iniFile = new IniFile( oldConnection.UDLPath );
            string connectionString = iniFile.ReadValue( "oledb", "Provider" );

            // Parse connection string
            ConnectionInfo connectionInfo = new ConnectionInfo();
            connectionInfo.Parse( connectionString );

            ConsolidateUdlConnection( connectionInfo, oldConnection, newConnection, oldConnection.ID, oldConnection.Name );
        }

        protected void ConsolidateUdlConnection( ConnectionInfo connectionInfo, DTS.Connection2 oldConnection, DTS.Connection2 newConnection, int id, string name ) {
            newConnection.DataSource = connectionInfo.DataSource;
            newConnection.Catalog = connectionInfo.InitialCatalog;
            newConnection.ConnectionProperties.Item( "Integrated Security" ).Value = connectionInfo.IntegratedSecurity;
            newConnection.ConnectionProperties.Item( "Persist Security Info" ).Value = connectionInfo.PersistSecurityInfo;
            newConnection.ConnectionProperties.Item( "Initial Catalog" ).Value = connectionInfo.InitialCatalog;
            newConnection.ConnectionProperties.Item( "Data Source" ).Value = connectionInfo.DataSource;
            if (connectionInfo.LocaleIdentifier != 0) 
                newConnection.ConnectionProperties.Item( "Locale Identifier" ).Value = connectionInfo.LocaleIdentifier;
            if (connectionInfo.PacketSize != 0)
                newConnection.ConnectionProperties.Item( "Packet Size" ).Value = connectionInfo.PacketSize;
            if (connectionInfo.ApplicationName.Length > 0)
                newConnection.ConnectionProperties.Item( "Application Name" ).Value = connectionInfo.ApplicationName;
            if (connectionInfo.WorkstationID.Length > 0)
                newConnection.ConnectionProperties.Item( "Workstation ID" ).Value = connectionInfo.WorkstationID;

            newConnection.Name = name;
            newConnection.ID = id;
            newConnection.Reusable = true; //TODO: optimize multiple connections referencing the same database
            newConnection.ConnectImmediate = oldConnection.ConnectImmediate;
            newConnection.DataSource = connectionInfo.DataSource;
            newConnection.ConnectionTimeout = (connectionInfo.ConnectionTimeout > 0) ? connectionInfo.ConnectionTimeout : connectionInfo.ConnectTimeout;
            newConnection.Catalog = connectionInfo.InitialCatalog;
            newConnection.UseTrustedConnection = (connectionInfo.Trusted_Connection == ConnectionInfo.IntegratedSecurityArguments.True)
                                                 || (connectionInfo.Trusted_Connection == ConnectionInfo.IntegratedSecurityArguments.Yes)
                                                 || (connectionInfo.Trusted_Connection == ConnectionInfo.IntegratedSecurityArguments.SSPI)
                                                 || (connectionInfo.IntegratedSecurity == ConnectionInfo.IntegratedSecurityArguments.True )
                                                 || (connectionInfo.IntegratedSecurity == ConnectionInfo.IntegratedSecurityArguments.Yes)
                                                 || (connectionInfo.IntegratedSecurity == ConnectionInfo.IntegratedSecurityArguments.SSPI);
            newConnection.UseDSL = false;
        }

        protected enum MigrationMessage {
            Error,
            Warning,
            Info
        }
        
        /// <summary>
        /// Log information on migration
        /// </summary>
        /// <param name="s">Message to log</param>
        protected static void Log( MigrationMessage severity, string s ) {
            Console.Write( severity.ToString().ToUpper() );
            Console.Write( ": " );
            Console.WriteLine( s );
        }
        
    }
}
